### Persefone.jl - a model of agricultural landscapes and ecosystems in Europe.
###
### This file reads in weather data for the selected location and makes it available
### to the rest of the model.
###

"""
    Weather

Holds the weather information for the whole simulation period.
"""
@kwdef struct Weather
    firstdate::Date
    lastdate::Date
    # TODO: types for some of these fields can be just `Vector{Float64}` after
    #       fixing missing measurement values (replacing with values
    #       approximated from previous/later measurements)
    windspeed::Vector{Union{Missing,Float64}}
    precipitation::Vector{Float64}
    sunshine::Vector{Union{Missing,Float64}}
    cloudcover::Vector{Union{Missing,Float64}}
    humidity::Vector{Union{Missing,Float64}}
    meantemp::Vector{Float64}
    maxtemp::Vector{Float64}
    mintemp::Vector{Float64}
    evapotranspiration::Vector{Float64}
end

function Base.length(weather::Weather)
    return Dates.value(weather.lastdate - weather.firstdate) + 1
end

"""
    daynumber(weather, date)

Returns the number of days, counting `weather.firstdate` as day `1`.
"""
function daynumber(weather::Weather, date::Date)
    @assert weather.firstdate <= date <= weather.lastdate
    return Dates.value(date - weather.firstdate) + 1
end

"""
    findspans(predicate_fn, array) -> Vector{Tuple{Int, Int}}

Returns spans of indices in a 1-d `array` where a `predicate_fn`
returns `true`.  The spans are returned as a `Vector{Tuple{Int,
Int}}`, where each tuple is of the form `(start_index, end_index)`.
"""
function findspans(predicate_fn, arr::AbstractVector)
    spans = Tuple{Int, Int}[]
    startidx = nothing
    for i = firstindex(arr):lastindex(arr)
        if predicate_fn(arr[i])
            if startidx === nothing
                startidx = i
            end
        elseif startidx !== nothing
            push!(spans, (startidx, i - 1))
            startidx = nothing
        end
    end
    if startidx !== nothing
        push!(spans, (startidx, lastindex(arr)))
    end
    return spans
end

"""
    check_missing_weatherdata(dataframe)

Check the weather input data for missing values in columns where input
values are required.
"""
function check_missing_weatherdata(df::DataFrame)
    colnames = ["min_temperature", "max_temperature", "mean_temperature",
                "precipitation", "potential_evapotranspiration"]
    for colname in colnames
        spans = findspans(ismissing, getproperty(df, colname))
        # TODO: fix missing values and warn here instead of error
        if ! isempty(spans)
            error("Column $colname has missing values: $spans")
        end
    end
end

"""
    initweather(weatherfile, startdate, enddate)

Load a weather file, extract the values that are relevant to this model run
(specified by start and end dates), and return a dictionary of Weather objects
mapped to dates.

**Note:** This requires a weather file in the format produced by
  `data/regions/auxiliary/extract_weather_data.R`.
"""
function initweather(weatherfile::String, startdate::Date, enddate::Date)
    @debug "Initialising weather"
    if startdate > enddate
        error("Startdate is after enddate: $startdate > $enddate.")
    end
    Tfloat = Union{Missing, Float64}
    df = CSV.read(weatherfile, DataFrame;
                  missingstring="NA", dateformat="yyyy-mm-dd",
                  types=[Date, Tfloat, Tfloat, Tfloat, Tfloat, Tfloat,
                         Tfloat, Tfloat, Tfloat, Tfloat])
    needed_cols = [
        "date", "mean_windspeed", "precipitation", "sunshine_hours",
        "mean_cloud_cover", "mean_humidity", "mean_temperature",
        "max_temperature", "min_temperature", "potential_evapotranspiration"
    ]
    if ! issubset(needed_cols, names(df))
        error("Missing columns in weather data file: $(setdiff(needed_cols, names(df)))")
    end
    filter!(r -> r.date >= startdate && r.date <= enddate, df)
    sort!(df, :date)
    if startdate:enddate != df.date
        error("There are missing days (rows) in the weather data file."
              * " expected $(Dates.value(enddate - startdate) + 1), got $(nrow(df)))."
              * " Missing dates are $(setdiff(startdate:enddate, df.date)).")
    end
    check_missing_weatherdata(df)
    return Weather(; firstdate = minimum(df.date),
                   lastdate = maximum(df.date),
                   windspeed = df.mean_windspeed,
                   precipitation = df.precipitation,
                   sunshine = df.sunshine_hours,
                   cloudcover = df.mean_cloud_cover,
                   humidity = df.mean_humidity,
                   meantemp = df.mean_temperature,
                   maxtemp = df.max_temperature,
                   mintemp = df.min_temperature,
                   evapotranspiration = df.potential_evapotranspiration)
end

"""
    windspeed(weather, date)

Return the average windspeed in m/s on `date`.
"""
windspeed(weather::Weather, date::Date) =
    weather.windspeed[daynumber(weather, date)]

"""
    precipitation(weather, date)

Return the total precipitation in mm on `date`.
"""
precipitation(weather::Weather, date::Date) =
    weather.precipitation[daynumber(weather, date)]

"""
    sunshine(weather, date)

Return the sunshine duration in hours on `date`.
"""
sunshine(weather::Weather, date::Date) =
    weather.sunshine[daynumber(weather, date)]

"""
    humidity(weather, date)

Return today's average vapour pressure in %.
"""
humidity(weather::Weather, date::Date) =
    weather.humidity[daynumber(weather, date)]

"""
    meantemp(weather, date)

Return the mean temperature in °C on `date`.
"""
meantemp(weather::Weather, date::Date) =
    weather.meantemp[daynumber(weather, date)]

"""
    maxtemp(weather, date)

Return the maximum temperature in °C on `date`.
"""
maxtemp(weather::Weather, date::Date) =
    weather.maxtemp[daynumber(weather, date)]

"""
    mintemp(weather, date)

Return the minimum temperature in °C on `date`.
"""
mintemp(weather::Weather, date::Date) =
    weather.mintemp[daynumber(weather, date)]

"""
    evapotranspiration(weather, date)

Return today's potential evapotranspiration (ETo) on `date`.
"""
evapotranspiration(weather::Weather, date::Date) =
    weather.evapotranspiration[daynumber(weather, date)]
